/*****************************************************************
 * FILE: alignalign_space.c
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: 17 May 2007
 * Version:  1.0.0
 * 
 ****************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>       /* strlen, strcat  */
#include <ctype.h>        /* isspace         */
#include "alignalign.h"

Shape **Sf;                         /* dynamic programming table forward */
Shape **Sr;                         /* dynamic programming table reverse */
Shape  *t1f;
Shape  *t2f;
Shape  *t1r;
Shape  *t2r;
int script_i = 0;
int script_len = 0;


/*
 * alignalign_s: Implementation of aligning alignments algorithm
 *               that uses space linear in the number of input columns.
 *               (Although in the worst case it's exponential in the
 *               number of input rows.)
 */
int alignalign_space (char ***alignment_ptr) 
{
  int      i, j;
  aa_ushort *flat;        /* holds the flat shape representation */
  int       Aindex, Bindex;
  script_i = 0;
  
  aa_script = aa_malloc(aa_N * sizeof(unsigned char));
  script_len = aa_N; //aa_script will have to be at least this long; will realloc if necessary

  /* allocate dynamic programming table */
  Sf = aa_malloc((aa_N+1) * sizeof(Shape*));
  Sr = aa_malloc((aa_N+1) * sizeof(Shape*));

  for (j = 0; j <= aa_N; j++) { 
    Sf[j] = aa_malloc(aa_nodesize);
    Sr[j] = aa_malloc(aa_nodesize);
    NEXT(Sf[j]) = Sf[j];
    NEXT(Sr[j]) = Sr[j]; 
  }

  /* flat shape */
  flat = aa_malloc((aa_K+aa_L) * sizeof(aa_ushort));
  for (i = 0; i < aa_K+aa_L; i++) { flat[i] = 0; }

  t1f = aa_malloc(aa_nodesize);
  t2f = aa_malloc(aa_nodesize);
  t1r = aa_malloc(aa_nodesize);
  t2r = aa_malloc(aa_nodesize);
  NEXT(t1f) = t1f;
  NEXT(t2f) = t2f;
  NEXT(t1r) = t1r;
  NEXT(t2r) = t2r;

  /* */
  aa_ushort *alloc_catcher;
  alloc_catcher = RecursiveDPArrayPathFinder(0, aa_M - 1, 0, aa_N - 1, flat, flat);
  free(alloc_catcher);

  for (j = 0; j <= aa_N; j++) { 
  	aa_free(Sf[j], aa_nodesize);
    aa_free(Sr[j], aa_nodesize);
  }
  aa_free(Sf, (aa_N+1) * sizeof(Shape*));
  aa_free(Sr, (aa_N+1) * sizeof(Shape*));
  aa_free(flat, (aa_K+aa_L) * sizeof(aa_ushort));
  
  nodepoolfree(1);

  /* allocate alignment matrix (to store final result) */
  *alignment_ptr = aa_malloc((aa_K+aa_L) * sizeof(char *));
  for (i = 0; i < aa_K+aa_L; i++) {
    (*alignment_ptr)[i] = aa_malloc((script_i + 1) * sizeof(char));
  }

  Aindex = Bindex = 0;

  for (j = 0; j < script_i; j++) {
    switch (aa_script[j]) {
    case AA_INS:
      for (i=0; i<aa_K; i++) { (*alignment_ptr)[i][j] = AA_EPS; }
      for (i=0; i<aa_L; i++) { (*alignment_ptr)[i+aa_K][j] = aa_B[i][Bindex]; } 
      Bindex++;
      //fprintf(stderr, "-");
      break;
      
    case AA_DEL:
      for (i=0; i<aa_K; i++) { (*alignment_ptr)[i][j] = aa_A[i][Aindex]; }
      Aindex++;
      for (i=0; i<aa_L; i++) { (*alignment_ptr)[i+aa_K][j] = AA_EPS; }
      //fprintf(stderr, "|");
      break;
      
    case AA_SUB:
      for (i=0; i<aa_K; i++) { (*alignment_ptr)[i][j] = aa_A[i][Aindex]; }
      Aindex++;
      for (i=0; i<aa_L; i++) { (*alignment_ptr)[i+aa_K][j] = aa_B[i][Bindex]; }
      Bindex++;
      //fprintf(stderr, "\\");
      break;
    };
  }
  //printf("\n");


  return script_i;
}


/********************************************************************
 * RecursiveDPArrayPathFinder
 *******************************************************************/
aa_ushort *RecursiveDPArrayPathFinder (int a1, int a2, int b1, int b2, 
		       aa_ushort *pred, aa_ushort *succ) {
  int m = a2 - a1 + 1, n = b2 - b1 + 1;
  int j;
  int new_a2;
  int new_b2 = -1;
  int min;
  Shape *p, *q, *header, *header2, *pmin = NULL, *qmin = NULL;
  aa_ushort *new_succ, *new_pred, *end;


  for (j = 0; j <= aa_N; j++) {
  	//free shapes of forward table
    header = q = Sf[j];
    p = NEXT(q);
    while (p != header) {
      q = NEXT(p);
      nodefree(p);
      p = q;
    }
    NEXT(header) = header;

  	//free shapes of reverse table    
  	header = q = Sr[j];
    p = NEXT(q);
    while (p != header) {
      q = NEXT(p);
      nodefree(p);
      p = q;
    }
    NEXT(header) = header;
  }

  if ( (m <= 1) || (n <= 0) || ((n <= 1) && (m <= aa_N))) {
    return BaseCase(a1, a2, b1, b2, pred, succ);
  }

  new_a2 = m/2 + a1 - 1; // m/2 is equal to floor(m/2) 


  EvaluateForward(a1, new_a2, b1, b2, pred);
  EvaluateReverse(new_a2 + 1, a2, b1, b2, succ);
  
  min = AA_INF;
  for (j = 0; j <= n; j++) {
    header = Sf[j];
    p = NEXT(header);
    while (p != header) {
      header2 = Sr[n-j];
      q = NEXT(header2);
      while (q != header2) {
/*      	if (a1 == 42 && (new_a2 == 51)) {
    		  fprintf(stderr, "j=%d: Cp=%d, Cq=%d \n", (b1+j),	C(p), C(q) );
			  fprintf(stderr, "\tp: ");
			  int ii;
      		  for (ii = 0; ii < aa_K + aa_L; ii++) { 
      		  	fprintf(stderr, "%d, ", S(p)[ii]);
      		  }
			  fprintf(stderr, "\n\tq: ");      		  
      		  for (ii = 0; ii < aa_K + aa_L; ii++) { 
      		  	fprintf(stderr, "%d, ", S(q)[ii]);
      		  }			  
    		  //fprintf(stderr, "\n\t%%d+ %d - (%d*%d) = %d\n", C(p), C(q), aa_gamma, compshapes(S(p), S(q)), (C(p) + C(q) - aa_gamma * compshapes(S(p), S(q))));
      	}*/
       	if ((int)(C(p) + C(q) - aa_gamma * compshapes(S(p), S(q))) < min) {
		  min = (int)(C(p) + C(q) - aa_gamma * compshapes(S(p), S(q)));
		  new_b2 = j + b1 - 1;
		  pmin = p;
		  qmin = q;
		}
		q = NEXT(q);
      }      
      p = NEXT(p);
    }
  }

//  fprintf(stderr, "****** min %u a1 %d, new_a2 %d, a2 %d; b1 %d, new_b2 %d, b2 %d ******\n", min, a1, new_a2, a2, b1, new_b2, b2);
  new_succ = aa_malloc((aa_K+aa_L) * sizeof(aa_ushort));
  for (j = 0; j < aa_K + aa_L; j++) { new_succ[j] = S(qmin)[j]; }
  
//  fprintf(stderr, "----- m,n (%d,%d) forward\n", new_a2-a1+1, new_b2-b1+1);
//  fprintf(stderr, "----- m,n (%d,%d) reverse\n", a2-new_a2,   b2-new_b2);

  new_pred = RecursiveDPArrayPathFinder(a1, new_a2, b1, new_b2, pred, new_succ);
  free(new_succ);

  end = RecursiveDPArrayPathFinder(new_a2 + 1, a2, new_b2 + 1, b2, new_pred, succ);
  free(new_pred);
  return end;
}


/********************************************************************
 * BaseCase
 *******************************************************************/
aa_ushort *BaseCase (int a1, int a2, int b1, int b2,
		   aa_ushort *pred, aa_ushort *succ) {
  static int i, j, min_index;
  static int min, cf, cf2;
  static aa_ushort sfor[AA_MAXSEQNUM];
  int m = a2 - a1 + 1;
  int n = b2 - b1 + 1;
  aa_ushort *end;

//  fprintf(stderr, "basis a1 %d a2 %d b1 %d b2 %d\n", a1,a2,b1,b2);
//  fprintf(stderr, "basis %d, %d\n", m,n);
  end = aa_malloc((aa_K+aa_L) * sizeof(aa_ushort));

  //m always >= 1, because of the conditions required to call this code
  

  if (n <= 0) {
    for (i = 0; i < aa_K + aa_L; i++) { end[i] = pred[i]; }
    for (i = a1; i <= a2; i++) {
      newshape(aa_A, NULL, i, -1, end, end);
      ensureScriptLongEnough();
      aa_script[script_i] = AA_DEL;
      script_i++;
    }
  }
  else if (m == 1) { 
    if (n == 1) {  //  --> substitution is assumed here
      newshape(aa_A, aa_B, a1, b1, pred, end);
      ensureScriptLongEnough();
      aa_script[script_i] = AA_SUB;
      script_i++;
    }
    else { // m == 1 and n > 1 (@@@ assumes 1 substitution - determines where)
      C(Sf[0]) = 0;
      C(Sr[0]) = 0;
      memcpy((void *) S(Sf[0]), (void *) pred, sizeof(aa_ushort) * (aa_K+aa_L));
      memcpy((void *) S(Sr[0]), (void *) succ, sizeof(aa_ushort) * (aa_K+aa_L));
      for (j = 0; j < n; j++) {
		newshape(NULL, aa_B, -1, b1+j, S(Sf[j]), S(Sf[j+1]));
		newshape(NULL, aa_B, -1, b2-j, S(Sr[j]), S(Sr[j+1]));
		C(Sf[j+1]) = C(Sf[j]) + subs(a1,b1+j,AA_HORIZ) + gaps(a1,b1+j,S(Sf[j]),AA_HORIZ);
		C(Sr[j+1]) = C(Sr[j]) + subs(a2,b2-j,AA_HORIZ) + gaps(a2,b2-j,S(Sr[j]),AA_HORIZ);
      }
      min = AA_INF;
      min_index = 0;
      for (j = 0; j < n; j++) {
		newshape(aa_A, aa_B, a1, b1+j, S(Sf[j]), sfor);
		cf = C(Sf[j]) + subs(a1,b1+j,AA_DIAG) + gaps(a1,b1+j,S(Sf[j]),AA_DIAG);
		if (cf + C(Sr[n-j-1]) - aa_gamma * compshapes(sfor, S(Sr[n-j-1]))  < min) {
		  min_index = j;
		  min = cf + C(Sr[n-j-1]) - aa_gamma * compshapes(sfor, S(Sr[n-j-1]));
		}
      }
      for (j = 0; j < min_index; j++) {
      	ensureScriptLongEnough();
		aa_script[script_i] = AA_INS;
		script_i++;
      }
   	  ensureScriptLongEnough();
      aa_script[script_i] = AA_SUB;
      script_i++;
      newshape(aa_A, aa_B, a1, b1+j, S(Sf[j]), end);
      for (j = min_index + 1; j < n; j++) {
      	ensureScriptLongEnough();      	
		aa_script[script_i] = AA_INS;
		script_i++;
		newshape(NULL, aa_B, -1, b1+j, end, end);
      }
    }
  }
  else if (n == 1) { // and m > 1 (@@@ assumes 1 substitution) 
    C(Sf[0]) = 0;
    C(Sr[0]) = 0;
    memcpy((void *) S(Sf[0]), (void *) pred, sizeof(aa_ushort) * (aa_K+aa_L));
    memcpy((void *) S(Sr[0]), (void *) succ, sizeof(aa_ushort) * (aa_K+aa_L));

    for (i = 0; i < m; i++) {
      newshape(aa_A, NULL, a1+i, -1, S(Sf[i]), S(Sf[i+1]));
      newshape(aa_A, NULL, a2-i, -1, S(Sr[i]), S(Sr[i+1]));
      C(Sf[i+1]) = C(Sf[i]) + subs(a1+i,b1,AA_VERT) + gaps(a1+i,b1,S(Sf[i]),AA_VERT);
      C(Sr[i+1]) = C(Sr[i]) + subs(a2-i,b2,AA_VERT) + gaps(a2-i,b2,S(Sr[i]),AA_VERT);
    }
    
    min = AA_INF;
    min_index = 0;

    for (i = 0; i < m; i++) {
      newshape(aa_A, aa_B, a1+i, b1, S(Sf[i]), sfor);
      cf = C(Sf[i]) + subs(a1+i,b1,AA_DIAG) + gaps(a1+i,b1,S(Sf[i]),AA_DIAG); // cost to make substitution
//	  This causes a bug ... but the replacement code below doesn't. Does that mean there's an invalid memory hit somewhere else?
/*      if (cf + C(Sr[m-i-1]) - aa_gamma * compshapes(sfor, S(Sr[m-i-1])) < min) {
		min_index = i;
		min = cf + C(Sr[m-i-1]) - aa_gamma * compshapes(sfor, S(Sr[m-i-1]));
      }
*/
      cf2 = cf + C(Sr[m-i-1]) - aa_gamma * compshapes(sfor, S(Sr[m-i-1])); // cost to stitch on the reverse
      if (cf2 < min) {
		min_index = i;
		min = cf2;
      }
    }
    for (i = 0; i < min_index; i++) {
      ensureScriptLongEnough();    	
      aa_script[script_i] = AA_DEL;
      script_i++;
    }
   	ensureScriptLongEnough();    
   	aa_script[script_i] = AA_SUB;
    script_i++;
    newshape(aa_A, aa_B, a1+i, b1, S(Sf[i]), end);
    for (i = min_index + 1; i < m; i++) {
      ensureScriptLongEnough();      
      aa_script[script_i] = AA_DEL;
      script_i++;

      newshape(aa_A, NULL, a1+i, -1, end, end);      
    }
  }
  else {
    fprintf(stderr, "*** ERROR in base case *** aborting...\n");
    exit(-1);
  }
    
  return end;
}


/********************************************************************
 * EvaluateForward 
 *******************************************************************/
void EvaluateForward (int a1, int a2, int b1, int b2, aa_ushort *pred) {
  static int       i, j, m, n;
  static aa_uint   c;
  static aa_ushort s[AA_MAXSEQNUM];
  static Shape    *p;
  static Shape    *q;
  static Shape    *header;
  static Shape    *t3;
  static int       prune;  

  m = a2 - a1 + 1;
  n = b2 - b1 + 1;
   
  p           = nodemalloc();
  NEXT(Sf[0]) = p;
  NEXT(p)     = Sf[0];
  C(p)        = 0;
  //for (j = 0; j < aa_K + aa_L; j++) { (S(p))[j] = pred[j]; }
  memcpy((void *) S(p), (void *) pred, sizeof(aa_ushort) * (aa_K+aa_L));

  /* fill in the dynamic programming table (row major order) */
  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++) {
      header = Sf[j];
      p = NEXT(header);
      while (p != header) {
		newshape(aa_A, NULL, i+a1, -1, S(p), s);
		c = C(p) + subs(i+a1,b1+j,AA_VERT) + gaps(i+a1,b1+j,S(p),AA_VERT);
		prune = dominance(t1f, s, c, i, j);
		if (!prune) { insert_space(t1f, s, c); }
		newshape(NULL, aa_B, -1, j+b1, S(p), s);
		c = C(p) + subs(a1+i,j+b1,AA_HORIZ) + gaps(a1+i,j+b1,S(p),AA_HORIZ);
		prune = dominance(Sf[j+1], s, c, i, j);
		if (!prune) { insert_space(Sf[j+1], s, c); }
		newshape(aa_A, aa_B, i+a1, j+b1, S(p), s);
		c = C(p) + subs(i+a1,j+b1,AA_DIAG) + gaps(i+a1,j+b1,S(p),AA_DIAG);
		prune = dominance(t2f, s, c, i, j);
		if (!prune) { insert_space(t2f, s, c); }
		p = NEXT(p);
      }
      t3 = Sf[j];
      p = NEXT(t3);
      while (p != t3) {	q = NEXT(p); nodefree(p); p = q; }
      NEXT(t3) = t3; // t3 empty
      Sf[j] = t1f;   // Sf[j] inserted with (i+1,j)
      t1f   = t2f;   // t1 inserted with (i+1,j+1)
      t2f   = t3;    // t2 empty
      t3    = NULL;
    }

    /*********** j == n ***********/
    header = Sf[j];
    p = NEXT(header);
    while (p != header) {
      newshape(aa_A, NULL, i+a1, -1, S(p), s);
      c = C(p) + subs(i+a1,b2,AA_VERT) + gaps(i+a1,b2,S(p),AA_VERT);
      prune = dominance(t1f, s, c, i, j);
      if (!prune) { insert_space(t1f, s, c); }
      p = NEXT(p);
    }
    t3 = Sf[j];
    p = NEXT(t3);
    while (p != t3) { q = NEXT(p); nodefree(p); p = q; }
    NEXT(t3) = t3; // t3 empty
    Sf[j] = t1f;   // Sf[n] inserted with (i+1,n) 
    t1f   = t3;    // t1 empty  (t2 already empty)
    t3    = NULL;
  }
  
  /************** i == m **************/
  for (j = 0; j < n; j++) {
    header = Sf[j];
    p = NEXT(header);
    while (p != header) {
      newshape(NULL, aa_B, -1, j+b1, S(p), s);
      c = C(p) + subs(a2,j+b1,AA_HORIZ) + gaps(a2,j+b1,S(p),AA_HORIZ);
      prune = dominance(Sf[j+1], s, c, i, j);
      if (!prune) { insert_space(Sf[j+1], s, c); }
      p = NEXT(p);
    }
  }

  return;
}

/********************************************************************
 * EvaluateReverse
 *******************************************************************/
void EvaluateReverse (int a1, int a2, int b1, int b2, aa_ushort *succ) {
  static int    i, j, m, n;
  static aa_uint  c;
  static aa_ushort s[AA_MAXSEQNUM];
  static Shape *p;
  static Shape *q;
  static Shape *header;
  static Shape *t3;
  static int    prune;

  m = a2 - a1 + 1;
  n = b2 - b1 + 1;
   
  p           = nodemalloc();
  NEXT(Sr[0]) = p;
  NEXT(p)     = Sr[0];
  C(p)        = 0;
  //for (j = 0; j < aa_K + aa_L; j++) { (S(p))[j] = succ[j]; }
  memcpy((void *) S(p), (void *) succ, sizeof(aa_ushort) * (aa_K+aa_L));

  /* fill in the dynamic programming table (row major order) */
  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++) {
      header = Sr[j];
      p = NEXT(header);
      while (p != header) {
		newshape(aa_A, NULL, a2-i, -1, S(p), s);
		c = C(p) + subs(a2-i,b2-j,AA_HORIZ) + gaps(a2-i,b2-j,S(p),AA_HORIZ);
		prune = dominance(t1r, s, c, i, j);
		if (!prune) { insert_space(t1r, s, c); }
		newshape(NULL, aa_B, -1, b2-j, S(p), s);
		c = C(p) + subs(a2-i,b2-j,AA_VERT) + gaps(a2-i,b2-j,S(p),AA_VERT);
		prune = dominance(Sr[j+1], s, c, i, j);
		if (!prune) { insert_space(Sr[j+1], s, c); }
		newshape(aa_A, aa_B, a2-i, b2-j, S(p), s);
		c = C(p) + subs(a2-i,b2-j,AA_DIAG) + gaps(a2-i,b2-j,S(p),AA_DIAG);
		prune = dominance(t2r, s, c, i, j);
		if (!prune) { insert_space(t2r, s, c); }
		p = NEXT(p);
      }
      t3 = Sr[j];
      p = NEXT(t3);
      while (p != t3) {	q = NEXT(p); nodefree(p); p = q; }
      NEXT(t3) = t3; // t3 empty
      Sr[j] = t1r;   // Sr[j] inserted with (i+1,j)
      t1r   = t2r;   // t1 inserted with (i+1,j+1)
      t2r   = t3;    // t2 empty
      t3    = NULL;
    }

    /*********** j == n ***********/
    header = Sr[j];
    p = NEXT(header);
    while (p != header) {
      newshape(aa_A, NULL, a2-i, -1, S(p), s);
      c = C(p) + subs(a2-i,b1,AA_VERT) + gaps(a2-i,b1,S(p),AA_VERT);
      prune = dominance(t1r, s, c, i, j);
      if (!prune) { insert_space(t1r, s, c); }
      p = NEXT(p);
    }
    t3 = Sr[j];
    p = NEXT(t3);
    while (p != t3) { q = NEXT(p); nodefree(p); p = q; }
    NEXT(t3) = t3; // t3 empty
    Sr[j] = t1r;   // Sr[n] inserted with (i+1,n) 
    t1r   = t3;    // t1 empty  (t2 already empty)
    t3    = NULL;
  }
  
  /************** i == m **************/
  for (j = 0; j < n; j++) {
    header = Sr[j];
    p = NEXT(header);
    while (p != header) {
      newshape(NULL, aa_B, -1, b2-j, S(p), s);
      c = C(p) + subs(a1,b2-j,AA_HORIZ) + gaps(a1,b2-j,S(p),AA_HORIZ);
      prune = dominance(Sr[j+1], s, c, i, j);
      if (!prune) { insert_space(Sr[j+1], s, c); }
      p = NEXT(p);
    }
  }

  return;
}

void ensureScriptLongEnough () {
	if (script_i >= script_len) {
		script_len += AA_SCRIPTALLOCSIZE;
		aa_script = aa_realloc(aa_script, script_len);
  	}
}
