/*****************************************************************
 * FILE: subs.c
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: 17 May 2007
 * Version:  1.0.0
 * 
 ****************************************************************/


#include <stdio.h>
#include <stdlib.h> /* NULL */
#include <string.h> /* memset */
#include "subs.h"
#include "alignalign.h"

unsigned char **aa_fa; // aa_fa[i][j] #times letter j occurs in column i of A
unsigned char **aa_ca; // aa_ca[i][j] what is letter j in column i

unsigned char **aa_fb;  // same
unsigned char **aa_cb;

/*
 * procedure: subs
 *   Input:
 *     aa_A, aa_B: multiple sequence alignments
 *     a, b: column indices into aa_A and aa_B respectively
 *     k, l: numbers of sequences in aa_A and aa_B respectively
 *        d: substitution aa_cost matrix
 *   Output: sc, a int representing the sum of pairwise
 *           substitution costs between aa_A and aa_B (not within aa_A and aa_B).
 *
 * This version computes substitution costs from preprocessed letter
 * frequency tables.  Time per call is O(sigma*sigma), where sigma is
 * the size of the alphabet.
 */
inline int
subs (int a, int b, int dir) {
	if (aa_weighted || aa_gamma != aa_gamma_terminal  || aa_lambda != aa_lambda_terminal) {		
		return subsWeighted(a,b,dir);
	} else {
		return subsUnweighted(a,b,dir);
	} 
}

inline int
subsUnweighted (int a, int b, int dir) {
  static int i, j;
  static int sc; //, sc2;

//  sc2 = subs2(a,b);

  sc = 0;

  if (/*0 > a*/ AA_HORIZ == dir) {
    j = 0;
    while (0 < aa_fb[b][j]) {    	
      sc += aa_lambda * aa_K * aa_fb[b][j];
      j++;
    }
  }
  else if (/*0 > b*/ AA_VERT == dir) {
    j = 0;
    while (0 < aa_fa[a][j]) {
      sc += aa_lambda  * aa_L * aa_fa[a][j];
      j++;
    }
  }
  else { //dir == AA_DIAG
    i = 0;
    int seen_a = 0;
    int seen_b = 0;
    while (0 < aa_fa[a][i]) {
      j = 0;
      seen_a += aa_fa[a][i];
      seen_b = 0;
      while (0 < aa_fb[b][j]) {
		sc += (aa_cost[aa_alphaconv[(int)aa_ca[a][i]]][aa_alphaconv[(int)aa_cb[b][j]]]) * aa_fa[a][i] * aa_fb[b][j];
      	seen_b += aa_fb[b][j];
        j++;
      }
      sc += aa_lambda * (aa_L - seen_b) * aa_fa[a][i];
      i++;
    }
    sc += aa_lambda * (aa_K - seen_a) * seen_b;
  }
  
  return sc;
}


/*
 * Compute the pairwise substitution aa_cost on column a of aa_A and b of aa_B 
 * (quadratic time version)
 */
inline int
subsWeighted (int a, int b, int dir) {
  static int i, j;
  static int sc;
  sc = 0.0;
  if (/*0 > a*/ AA_HORIZ == dir) {
    for (j = 0; j < aa_L; j++) {
		if (AA_EPS != aa_B[j][b] ) {
	    	for (i=0; i<aa_K; i++) {
	    		sc += aa_weights[i][aa_K+j] * GET_LAMBDA_A(i,a-1);  
	    	}
		}
    }
  }
  else if (/*0 > b*/ AA_VERT == dir) {
    for (i = 0; i < aa_K; i++) {
		if (AA_EPS != aa_A[i][a]) {
	    	for (j=0; j<aa_L; j++) {
	    		sc += aa_weights[i][aa_K+j] * GET_LAMBDA_B(j,b-1);
	    	} 
		}
    }
  }
  else { //if (a >= 0 && b >= 0) {
    for (i = 0; i < aa_K; i++) {
      for (j = 0; j < aa_L; j++) {
      	if (AA_EPS != aa_A[i][a] && AA_EPS != aa_B[j][b] ) {
			sc += aa_cost[aa_alphaconv[(int)aa_A[i][a]]][aa_alphaconv[(int)aa_B[j][b]]] * aa_weights[i][aa_K+j];
      	} else if (AA_EPS != aa_A[i][a]) { 
			sc += aa_weights[i][aa_K+j] * GET_LAMBDA_B(j,b);  
      	} else if (AA_EPS != aa_B[j][b] ) {
			sc += aa_weights[i][aa_K+j] * GET_LAMBDA_A(i,a);
      	}
      }
    }
  }

  return sc;
}

/* initialize letter frequency table for substitution cost computations */
void init_frequencies(void) {
  int i, j, c;
  unsigned char tempfreq[AA_MAXALPHA];

  aa_fa = aa_malloc((unsigned int) ((aa_M + 1) * sizeof(unsigned char *)));
  aa_fb = aa_malloc((unsigned int) ((aa_N + 1) * sizeof(unsigned char *)));
  aa_ca = aa_malloc((unsigned int) ((aa_M + 1) * sizeof(unsigned char *)));
  aa_cb = aa_malloc((unsigned int) ((aa_N + 1) * sizeof(unsigned char *)));

  for (j = 0; j < aa_M; j++) {
    c = 0;
    memset(tempfreq, 0, AA_MAXALPHA * sizeof(unsigned char));
    for (i = 0; i < aa_K; i++) {
      if (aa_A[i][j] != AA_EPS){		
	      tempfreq[ aa_alphaconv[ (int)aa_A[i][j] ] ] += 1; //tempfreq[i] is # times letter i occurs in column
      }
    }
    for (i = 0; i < aa_alphasize; i++) {
      if (0 != tempfreq[i]) { c++; }  //c is # letters represented in current column
    }
    aa_fa[j] = aa_malloc(((unsigned int) c + 1) * sizeof(unsigned char));
    aa_ca[j] = aa_malloc(((unsigned int) c ) * sizeof(unsigned char));
    //memset(aa_fa[j], 0, c+1);
    c = 0;
    for (i = 0; i < aa_alphasize; i++) {
      if (0 != tempfreq[i]) {
		aa_fa[j][c] = tempfreq[i];     // #times letter c occurs in column j of A
		aa_ca[j][c] = aa_alphabet[i];  // what is letter c in column j
		c++;
      }
    }
    aa_fa[j][c] = 0;
  }
  for (j = 0; j < aa_N; j++) {
    c = 0;
    memset(tempfreq, 0, AA_MAXALPHA * sizeof(unsigned char));
    for (i = 0; i < aa_L; i++) {
      if (aa_B[i][j] != AA_EPS){		
	      tempfreq[ aa_alphaconv[ (int) aa_B[i][j] ] ] += 1;
      }
    }
    for (i = 0; i < aa_alphasize; i++) {
      if (0 != tempfreq[i]) { c++; }
    }
    aa_fb[j] = aa_malloc(((unsigned int) c + 1) * sizeof(unsigned char));
    aa_cb[j] = aa_malloc(((unsigned int) c ) * sizeof(unsigned char));
    //memset(aa_fb[j], 0, c+1); 
    c = 0;
    for (i = 0; i < aa_alphasize; i++) {
      if (0 != tempfreq[i]) {
		aa_fb[j][c] = tempfreq[i];
		aa_cb[j][c] = aa_alphabet[i];
		c++;
      }
    }
    aa_fb[j][c] = 0;
  }

  /*
  for (j = 0; j < aa_M; j++) {
    i = 0;
    fprintf(stderr, "%d ", j);
    while (aa_fa[j][i] > 0) {
      fprintf(stderr, "[%3d %c] ", aa_fa[j][i], aa_ca[j][i]);
      i++;
    }
    fprintf(stderr, "\n");
  }
  */
}

void free_frequencies(void) {
  int j;	
  for (j = 0; j < aa_M; j++) {
    free(aa_fa[j]);
    free(aa_ca[j]);
  }
  for (j = 0; j < aa_N; j++) {
    free(aa_fb[j]);
    free(aa_cb[j]);
  }
  free(aa_fa);
  free(aa_fb);
  free(aa_ca);
  free(aa_cb); 
}
